/*
Copyright (C) 2004-2008 Grame

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

*/

#ifndef __JackSocketClientChannel__
#define __JackSocketClientChannel__

#include "JackGenericClientChannel.h"
#include "JackSocket.h"
#include "JackPlatformPlug.h"
#include "JackThread.h"

namespace Jack
{

class JackClient;

/*!
\brief JackClientChannel using sockets.
*/

class JackSocketClientChannel : public JackGenericClientChannel, public JackRunnableInterface
{

    private:

        JackServerSocket fNotificationListenSocket; // Socket listener for server notification
        JackClientSocket* fNotificationSocket;      // Socket for server notification
        JackThread fThread;                         // Thread to execute the event loop
        JackClient* fClient;

    public:

        JackSocketClientChannel();
        virtual ~JackSocketClientChannel();

        int Open(const char* server_name, const char* name, jack_uuid_t uuid, char* name_res, JackClient* client, jack_options_t options, jack_status_t* status);
        void Close();

        int Start();
        void Stop();

        // JackRunnableInterface interface
        bool Init();
        bool Execute();

        bool IsChannelThread() { return fThread.IsThread(); }
};

} // end of namespace

#endif

